# TinyTuya Setup Wizard
# -*- coding: utf-8 -*-
"""
TinyTuya Setup Wizard Tuya based WiFi smart devices

Author: Jason A. Cox
For more information see https://github.com/jasonacox/tinytuya

Description
    Setup Wizard will prompt user for Tuya IoT Developer credentials and will gather all of
    the Device IDs and their Local KEYs.  It will save the credentials and the device
    data in the tinytuya.json configuration file and optionally scan the local devices.

    HOW to set up your Tuya IoT Developer account: iot.tuya.com:
    https://github.com/jasonacox/tinytuya#get-the-tuya-device-local-key

    This standalone script is also available in the module: tinytuya.wizard()

Credits
* Tuya API Documentation
    https://developer.tuya.com/en/docs/iot/open-api/api-list/api?id=K989ru6gtvspg
* TuyaAPI https://github.com/codetheweb/tuyapi by codetheweb and blackrozes
    The TuyAPI/CLI wizard inspired and informed this python version.
"""
# Modules
import requests
import time
import hmac
import hashlib
import json
import pprint
import logging
import tinytuya

# Backward compatability for python2
try:
    input = raw_input
except NameError:
    pass

# Terminal Color Formatting
bold = "\033[0m\033[97m\033[1m"
subbold = "\033[0m\033[32m"
normal = "\033[97m\033[0m"
dim = "\033[0m\033[97m\033[2m"
alert = "\033[0m\033[91m\033[1m"
alertdim = "\033[0m\033[91m\033[2m"

# Get Configuration Data
CONFIGFILE = 'tinytuya.json'
DEVICEFILE = 'devices.json'
config = {}
config['apiKey'] = ''
config['apiSecret'] = ''
config['apiRegion'] = ''
config['apiDeviceID'] = ''
needconfigs = True
try:
    # Load defaults
    with open(CONFIGFILE) as f:
        config = json.load(f)
except:
    # First Time Setup
    pass

print(bold + 'TinyTuya Setup Wizard' + dim + ' [%s]' % (tinytuya.version) + normal)
print('')

if(config['apiKey'] != '' and config['apiSecret'] != '' and
    config['apiRegion'] != '' and config['apiDeviceID'] != ''):
    needconfigs = False
    print("    " + subbold + "Existing settings:" + dim +
        "\n        API Key=%s \n        Secret=%s\n        Region=%s\n        DeviceID=%s" %
        (config['apiKey'], config['apiSecret'], config['apiRegion'],
        config['apiDeviceID']))
    print('')
    answer = input(subbold + '    Use existing credentials ' +
                normal + '(Y/n): ')
    if(answer[0:1].lower() == 'n'):
        needconfigs = True

if(needconfigs):
    # Ask user for config settings
    print('')
    config['apiKey'] = input(subbold + "    Enter " + bold + "API Key" + subbold +
        " from tuya.com: " + normal)
    config['apiSecret'] = input(subbold + "    Enter " + bold + "API Secret" + subbold +
        " from tuya.com: " + normal)
    # TO DO - Determine apiRegion from Device (default = us)
    config['apiRegion'] = 'us'
    config['apiDeviceID'] = input(subbold +
        "    Enter " + bold + "any Device ID" + subbold +
        " currently registered in Tuya App (used to pull full list): " + normal)
    # Write Config
    json_object = json.dumps(config, indent=4)
    with open(CONFIGFILE, "w") as outfile:
        outfile.write(json_object)
    print(bold + "\n>> Configuration Data Saved to %s" % CONFIGFILE)
    print(dim + json_object)

KEY = config['apiKey']
SECRET = config['apiSecret']
DEVICEID = config['apiDeviceID']
REGION = config['apiRegion']        # us, eu, cn, in
LANG = 'en'                         # en or zh

# Build Tuya IoT Platform Payload
"""
Header Elements:
Parameter 	  Type    Required	Description
client_id	  String     Yes	client_id
signature     String     Yes	HMAC-SHA256 Signature (see below)
sign_method	  String	 Yes	Message-Digest Algorithm of the signature: HMAC-SHA256.
t	          Long	     Yes	13-bit standard timestamp (now in milliseconds).
lang	      String	 No	    Language. It is zh by default in China and en in other areas.
access_token  String     *      Required for service management calls

Signature Details:
* Token Management: signature = HMAC-SHA256(KEY + t, SECRET).toUpperCase()
* Service Management: signature = HMAC-SHA256(KEY + access_token + t, SECRET).toUpperCase()

URIs:
* Get Token = https://openapi.tuyaus.com/v1.0/token?grant_type=1
* Get UserID = https://openapi.tuyaus.com/v1.0/devices/{DeviceID}
* Get Devices = https://openapi.tuyaus.com/v1.0/users/{UserID}/devices

"""

url = "https://openapi.tuya%s.com/v1.0/" % REGION
now = int(time.time()*1000)
payload = KEY + str(now)

# Sign Payload
signature = hmac.new(
    SECRET.encode('utf-8'),
    msg=payload.encode('utf-8'),
    digestmod=hashlib.sha256
).hexdigest().upper()

# Create Header Data
headers = {}
headers['client_id'] = KEY
headers['sign_method'] = 'HMAC-SHA256'
headers['t'] = str(now)
headers['sign'] = signature

# Get Token
response = requests.get(url + 'token?grant_type=1', headers=headers)
try:
    response_dict = json.loads(response.content.decode())
except:
    print("Failed to get valide JSON response")

# Rebuild Payload with Token
token = response_dict['result']['access_token']
now = int(time.time()*1000)
payload = KEY + token + str(now)

# Sign Payload
signature = hmac.new(
    SECRET.encode('utf-8'),
    msg=payload.encode('utf-8'),
    digestmod=hashlib.sha256
).hexdigest().upper()

# Create Header Data
headers = {}
headers['client_id'] = KEY
headers['sign_method'] = 'HMAC-SHA256'
headers['t'] = str(now)
headers['sign'] = signature
headers['access_token'] = token

# Get UID from sample Device ID to Pull Full List
response = requests.get(url + 'devices/%s' % DEVICEID, headers=headers)
try:
    response_dict = json.loads(response.content.decode())
except:
    print("Failed to get valide JSON response")

# Build Payload to get Devices by UID
uid = response_dict['result']['uid']
now = int(time.time()*1000)
payload = KEY + token + str(now)

# Sign Payload
signature = hmac.new(
    SECRET.encode('utf-8'),
    msg=payload.encode('utf-8'),
    digestmod=hashlib.sha256
).hexdigest().upper()

# Create Header Data
headers = {}
headers['client_id'] = KEY
headers['sign_method'] = 'HMAC-SHA256'
headers['t'] = str(now)
headers['sign'] = signature
headers['access_token'] = token

# Get list of Devices
response = requests.get(url + 'users/%s/devices' % uid, headers=headers)
json_data = json.loads(response.content.decode())
# pprint.pprint(json_data)

# Filter to only Name, ID and Key
tuyadevices = []
for i in json_data['result']:
    item = {}
    item['name'] = i['name']
    item['id'] = i['id']
    item['key'] = i['local_key']
    tuyadevices.append(item)

# Display device list
print("\n\n" + bold + "Device Listing\n" + dim)
output = json.dumps(tuyadevices, indent=4)  # sort_keys=True)
print(output)

# Save list to devices.json
print(bold + "\n>> " + normal + "Saving list to devices.json")
with open(DEVICEFILE, "w") as outfile:
    outfile.write(output)
print(dim + "    %d registered devices saved" % len(tuyadevices))

# Find out if we should poll all devices
answer = input(subbold + '\nPoll local devices? ' +
        normal + '(Y/n): ')
if(answer[0:1].lower() != 'n'):
    # Scan network for devices and provide polling data
    print(normal + "\nScanning local network for Tuya devices...")
    devices = tinytuya.deviceScan(False, 20)
    print("    %s%s local devices discovered%s" % (dim, len(devices), normal))
    print()

    def getIP(d, gwid):
        for ip in d:
            if (gwid == d[ip]['gwId']):
                return (ip, d[ip]['version'])
        return (0, 0)

    print("Polling local devices...")
    for i in tuyadevices:
            name = i['name']
            (ip, ver) = getIP(devices, i['id'])
            if (ip == 0):
                print("    %s[%s] - %s%s - %sError: No IP found%s" %
                    (subbold, name, dim, ip, alert, normal))
            else:
                try:
                    d = tinytuya.OutletDevice(i['id'], ip, i['key'])
                    if ver == "3.3":
                        d.set_version(3.3)
                    data = d.status()
                    if data:
                        state = alertdim + "Off" + dim
                        # print(data)
                        try:
                            if(data['dps']['1'] == True):
                                state = bold + "On" + dim
                            print("    %s[%s] - %s%s - %s - DPS: %r" %
                                (subbold, name, dim, ip, state, data['dps']))
                        except:
                            print("    %s[%s] - %s%s - %sNo Response" %
                                (subbold, name, dim, ip, alertdim))
                    else:
                        print("    %s[%s] - %s%s - %sNo Response" %
                            (subbold, name, dim, ip, alertdim))
                except:
                    print("    %s[%s] - %s%s - %sNo Response" %
                        (subbold, name, dim, ip, alertdim))

print("\nDone.\n")